from __future__ import annotations

import shutil
from contextlib import contextmanager
from os import PathLike
from pathlib import Path
from typing import IO, Any, BinaryIO, ContextManager, Iterator, TextIO, overload

from minato.filesystems.filesystem import FileSystem
from minato.util import DecompressOption, OpenBinaryMode, OpenTextMode, remove_file_or_directory, xopen


@FileSystem.register(["file", "osfs", ""])
class OSFileSystem(FileSystem):
    def __init__(self, url_or_filename: str | PathLike) -> None:
        super().__init__(url_or_filename)
        self._path = Path(self._url.path)

    def exists(self) -> bool:
        return self._path.exists()

    def download(self, path: str | PathLike) -> None:
        shutil.copy(self._path, path)

    def delete(self) -> None:
        if not self._path.exists():
            raise FileNotFoundError(self._path)
        remove_file_or_directory(self._path)

    def get_version(self) -> str | None:
        return str(self._path.stat().st_mtime_ns)

    @overload
    def open_file(
        self,
        mode: OpenTextMode = ...,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[TextIO]: ...

    @overload
    def open_file(
        self,
        mode: OpenBinaryMode,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[BinaryIO]: ...

    @overload
    def open_file(
        self,
        mode: str,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        decompress: DecompressOption = ...,
    ) -> ContextManager[IO[Any]]: ...

    def open_file(
        self,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
        *,
        decompress: DecompressOption = "none",
    ) -> ContextManager[IO[Any]]:
        @contextmanager
        def _open(
            mode: str,
            buffering: int,
            encoding: str | None,
            errors: str | None,
            newline: str | None,
            decompress: DecompressOption,
        ) -> Iterator[IO[Any]]:
            with xopen(
                self._path,
                mode=mode,
                buffering=buffering,
                encoding=encoding,
                errors=errors,
                newline=newline,
                decompress=decompress,
            ) as fp:
                yield fp

        return _open(mode, buffering, encoding, errors, newline, decompress)



import argparse
import datetime
from pathlib import Path

from minato.commands.subcommand import Subcommand
from minato.common import Table
from minato.config import Config
from minato.minato import Minato
from minato.util import is_archive_file, sizeof_fmt


@Subcommand.register(
    name="list",
    description="show list of cached files",
)
class ListCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument(
            "query",
            nargs="*",
            default=[],
            help="query to filter cached files",
        )
        self.parser.add_argument(
            "--sort",
            type=str,
            default=None,
            help="sort by the specified key",
        )
        self.parser.add_argument(
            "--desc",
            action="store_true",
            help="sort the list in descending order",
        )
        self.parser.add_argument(
            "--details",
            action="store_true",
            help="show more detailed information",
        )
        self.parser.add_argument(
            "--column-width",
            type=int,
            default=None,
            help="specify the max column width",
        )
        self.parser.add_argument(
            "--root",
            type=Path,
            default=None,
            help="specify the root directory path of cached data",
        )
        self.parser.add_argument(
            "--expired",
            action="store_true",
            default=None,
            help="show cached files that were expired",
        )
        self.parser.add_argument(
            "--failed",
            action="store_true",
            default=None,
            help="show cached files that were failed to download",
        )
        self.parser.add_argument(
            "--completed",
            action="store_true",
            default=None,
            help="show cached files that were successfully downloaded",
        )
        self.parser.add_argument(
            "--no-shrink",
            action="store_true",
            help="disable to shrink column width",
        )

    def run(self, args: argparse.Namespace) -> None:
        config = Config.load(
            cache_root=args.root,
        )
        minato = Minato(config)
        cache = minato.cache

        cached_files = cache.filter(
            queries=args.query,
            expired=args.expired,
            failed=args.failed,
            completed=args.completed,
        )

        columns = ["uid", "url", "size", "type", "status", "expire_days"]
        if args.details:
            columns.append("local_path")
            columns.append("created_at")
            columns.append("updated_at")
            columns.append("extraction_path")

        table = Table(columns=columns, shrink=not args.no_shrink)

        for cached_file in cached_files:
            info = cached_file.to_dict()

            info["uid"] = info["uid"][:8]

            if cached_file.local_path.exists():
                info["size"] = sizeof_fmt(cached_file.local_path.stat().st_size)
            else:
                info["size"] = "-"

            if cache.is_expired(cached_file):
                info["expire_days"] = f"EXPIRED({cached_file.expire_days})"
            elif cached_file.expire_days < 0:
                info["expire_days"] = "NONE"
            else:
                now = datetime.datetime.now()
                delta = now - cached_file.updated_at
                info["expire_days"] = f"{delta.days}/{cached_file.expire_days}"

            info["type"] = get_cache_type(cached_file.local_path)

            table.add(info)

        if args.sort:
            table.sort(args.sort, args.desc)

        table.show()


def get_cache_type(path: Path) -> str:
    if path.is_dir():
        return "dir"
    if is_archive_file(path):
        return "archive"
    return "file"



from minato.filesystems.filesystem import (  # noqa: F401
    FileSystem,
    delete,
    download,
    exists,
    get_version,
    open_file,
    upload,
)
from minato.filesystems.gcs import GCSFileSystem  # noqa: F401
from minato.filesystems.hf import HuggingfaceHubFileSystem  # noqa: F401
from minato.filesystems.http import HttpFileSystem  # noqa: F401
from minato.filesystems.osfs import OSFileSystem  # noqa: F401
from minato.filesystems.s3 import S3FileSystem  # noqa: F401



from __future__ import annotations

import logging
from os import PathLike
from pathlib import Path
from typing import IO, Any, BinaryIO, ContextManager, TextIO, overload

from minato.cache import Cache, CacheStatus
from minato.config import Config
from minato.exceptions import CacheNotFoundError, InvalidCacheStatus
from minato.filesystems import delete, download, exists, get_version, open_file, upload
from minato.util import (
    DecompressOption,
    OpenBinaryMode,
    OpenTextMode,
    extract_archive_file,
    extract_path,
    is_archive_file,
    is_local,
    remove_file_or_directory,
)

logger = logging.getLogger(__name__)


class Minato:
    def __init__(self, config: Config | None = None) -> None:
        self._config = config or Config.load()
        self._cache = Cache(
            root=self._config.cache_root,
            default_expire_days=self._config.expire_days,
            default_auto_update=self._config.auto_update,
        )

    @property
    def cache(self) -> Cache:
        return self._cache

    @overload
    def open(
        self,
        url_or_filename: str | PathLike,
        mode: OpenTextMode = ...,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        extract: bool = ...,
        decompress: DecompressOption = ...,
        auto_update: bool | None = ...,
        expire_days: int | None = ...,
        use_cache: bool = ...,
        force_download: bool = ...,
        force_extract: bool = ...,
        retry: bool = ...,
    ) -> ContextManager[TextIO]: ...

    @overload
    def open(
        self,
        url_or_filename: str | PathLike,
        mode: OpenBinaryMode,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        extract: bool = ...,
        decompress: DecompressOption = ...,
        auto_update: bool | None = ...,
        expire_days: int | None = ...,
        use_cache: bool = ...,
        force_download: bool = ...,
        force_extract: bool = ...,
        retry: bool = ...,
    ) -> ContextManager[BinaryIO]: ...

    @overload
    def open(
        self,
        url_or_filename: str | PathLike,
        mode: str,
        buffering: int = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        newline: str | None = ...,
        *,
        extract: bool = ...,
        decompress: DecompressOption = ...,
        auto_update: bool | None = ...,
        expire_days: int | None = ...,
        use_cache: bool = ...,
        force_download: bool = ...,
        force_extract: bool = ...,
        retry: bool = ...,
    ) -> ContextManager[IO[Any]]: ...

    def open(
        self,
        url_or_filename: str | PathLike,
        mode: str = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
        *,
        extract: bool = False,
        decompress: DecompressOption = "none",
        auto_update: bool | None = None,
        expire_days: int | None = None,
        use_cache: bool = True,
        force_download: bool = False,
        force_extract: bool = False,
        retry: bool = True,
    ) -> ContextManager[IO[Any]]:
        if not ("a" in mode or "w" in mode or "x" in mode or "+" in mode) and use_cache:
            logger.debug("Open cached file of %s.", url_or_filename)
            url_or_filename = self.cached_path(
                url_or_filename,
                extract=extract,
                auto_update=auto_update,
                expire_days=expire_days,
                force_download=force_download,
                force_extract=force_extract,
                retry=retry,
            )

        return open_file(url_or_filename, mode, buffering, encoding, errors, newline, decompress=decompress)

    def cached_path(
        self,
        url_or_filename: str | PathLike,
        extract: bool = False,
        auto_update: bool | None = None,
        expire_days: int | None = None,
        force_download: bool = False,
        force_extract: bool = False,
        retry: bool = True,
    ) -> Path:
        url_or_filename = str(url_or_filename)

        if "!" in url_or_filename:
            remote_archive_path, file_path = url_or_filename.rsplit("!", 1)
            archive_path = self.cached_path(
                remote_archive_path,
                extract=True,
                force_extract=force_extract,
                force_download=force_download,
            )
            if not archive_path.is_dir():
                raise ValueError(f"{url_or_filename} uses the ! syntax, but this is not an archive file.")

            file_path = extract_path(file_path)
            local_path = archive_path / file_path
            if not local_path.exists():
                raise FileNotFoundError(local_path)
            return local_path

        if is_local(url_or_filename):
            url_or_filename = extract_path(url_or_filename)
            if not (extract and is_archive_file(url_or_filename)):
                local_path = Path(url_or_filename)
                if not local_path.exists():
                    raise FileNotFoundError(local_path)
                return local_path

        url = url_or_filename
        try:
            cached_file = self._cache.by_url(url)
        except CacheNotFoundError:
            cached_file = self._cache.new(url)

        with self._cache.lock(cached_file):
            if not self._cache.exists(cached_file):
                self._cache.add(cached_file)

            cached_file = self._cache.by_uid(cached_file.uid)

            if expire_days is not None:
                cached_file.expire_days = expire_days
                self._cache.save(cached_file)

            if auto_update is not None:
                cached_file.auto_update = auto_update
                self._cache.save(cached_file)

            if cached_file.auto_update and cached_file.version is not None:
                current_version = get_version(cached_file.url)
                if current_version != cached_file.version:
                    logger.debug("New version of data is available. It will be automatically updated.")
                    force_download = True

            if retry and cached_file.status != CacheStatus.COMPLETED:
                force_download = True

            try:
                downloaded = False
                if not cached_file.local_path.exists() or self._cache.is_expired(cached_file) or force_download:
                    remove_file_or_directory(cached_file.local_path)

                    cached_file.status = CacheStatus.DOWNLOADING
                    self._cache.update(cached_file)

                    logger.debug(
                        "Start downloading file(s) from %s to %s.",
                        cached_file.url,
                        cached_file.local_path,
                    )
                    self.download(cached_file.url, cached_file.local_path)
                    cached_file.version = get_version(cached_file.url)
                    downloaded = True

                extracted = False
                if (
                    (extract and cached_file.extraction_path is None)
                    or (downloaded and cached_file.extraction_path is not None)
                    or force_extract
                ) and is_archive_file(cached_file.local_path):
                    cached_file.extraction_path = Path(str(cached_file.local_path) + "-extracted")
                    remove_file_or_directory(cached_file.extraction_path)

                    cached_file.status = CacheStatus.EXTRACTING
                    self._cache.update(cached_file)

                    logging.debug(
                        "Extracting archive file from %s to %s.",
                        cached_file.local_path,
                        cached_file.extraction_path,
                    )
                    extract_archive_file(cached_file.local_path, cached_file.extraction_path)
                    extracted = True

                if downloaded or extracted:
                    cached_file.status = CacheStatus.COMPLETED
                    self._cache.update(cached_file)
            except FileNotFoundError:
                self._cache.delete(cached_file)
                raise
            except (Exception, SystemExit, KeyboardInterrupt):
                cached_file.status = CacheStatus.FAILED
                self._cache.update(cached_file)
                raise

        if (extract or force_extract) and cached_file.extraction_path:
            if not cached_file.extraction_path.exists():
                raise FileNotFoundError(cached_file.extraction_path)
            if cached_file.status != CacheStatus.COMPLETED:
                raise InvalidCacheStatus(f"Cached path status is not completed: status={cached_file.status}")
            return cached_file.extraction_path

        if not cached_file.local_path.exists():
            raise FileNotFoundError(cached_file.local_path)
        if cached_file.status != CacheStatus.COMPLETED:
            raise InvalidCacheStatus(f"Cached path status is not completed: status={cached_file.status}")
        return cached_file.local_path

    def available_update(self, url_or_filename: str | PathLike) -> bool:
        if is_local(url_or_filename):
            return False

        url = str(url_or_filename)
        cached_file = self._cache.by_url(url)
        current_version = get_version(url)
        return cached_file.version != current_version

    @staticmethod
    def download(url: str, filename: str | PathLike) -> None:
        download(url, filename)

    @staticmethod
    def upload(filename: str | PathLike, url: str) -> None:
        upload(filename, url)

    @staticmethod
    def delete(url_or_filename: str | PathLike) -> None:
        delete(url_or_filename)

    @staticmethod
    def exists(url_or_filename: str | PathLike) -> bool:
        return exists(url_or_filename)



from minato.common.filelock import FileLock  # noqa: F401
from minato.common.progress import Progress  # noqa: F401
from minato.common.selector import Selector  # noqa: F401
from minato.common.table import Table  # noqa: F401
from minato.common.url import URL  # noqa: F401



from __future__ import annotations

import os
import shutil
import subprocess

SELECTOR_COMMAND = os.environ.get("MINATO_SELECTOR_COMMAND")


class Selector:
    SUPPORTED_SELECTOR_COMMANDS = ["fzf", "peco"]

    def __init__(
        self,
        selector_command: str | None = None,
    ) -> None:
        self._selector_command = selector_command or self._find_selector_command()

    def __call__(self, items: list[str]) -> str | None:
        if self._selector_command:
            return self._select_with_command(items)
        return self._select_without_command(items)

    def _format_list(self, items: list[str]) -> str:
        return "\n".join(f"{i}: {item}" for i, item in enumerate(items, start=1))

    def _parse_result(self, line: str) -> str:
        return line.strip().split(": ", 1)[-1]

    def _select_without_command(self, items: list[str]) -> str | None:
        print(self._format_list(items))
        try:
            index = int(input("select index > ")) - 1
        except ValueError:
            index = -1
        if 0 <= index < len(items):
            return items[index]
        return None

    def _select_with_command(self, items: list[str]) -> str | None:
        assert self._selector_command is not None

        proc = subprocess.Popen(
            [self._selector_command],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=None,
        )
        stdin = proc.stdin
        stdout = proc.stdout

        assert stdin is not None
        assert stdout is not None

        stdin.write(self._format_list(items).encode())

        stdin.flush()
        stdin.close()
        proc.wait()

        result = stdout.read().decode().strip()
        if result:
            return self._parse_result(result)
        return None

    def _find_selector_command(self) -> str | None:
        if SELECTOR_COMMAND:
            return SELECTOR_COMMAND
        for cmd in self.SUPPORTED_SELECTOR_COMMANDS:
            if shutil.which(cmd):
                return cmd
        return None



from __future__ import annotations

import dataclasses
from configparser import ConfigParser
from os import PathLike
from pathlib import Path

MINATO_ROOT = Path.home() / ".minato"
DEFAULT_CACHE_ROOT = MINATO_ROOT / "cache"
ROOT_CONFIG_PATH = MINATO_ROOT / "config.ini"
LOCAL_CONFIG_PATH = Path.cwd() / "minato.ini"


@dataclasses.dataclass
class Config:
    cache_root: Path = DEFAULT_CACHE_ROOT
    expire_days: int = -1
    auto_update: bool = True
    selector_command: str | None = None

    @classmethod
    def load(
        cls,
        cache_root: str | PathLike | None = None,
        expire_days: int | None = None,
        auto_update: bool | None = None,
        files: list[str | PathLike] | None = None,
    ) -> Config:
        if files is None:
            files = [ROOT_CONFIG_PATH, LOCAL_CONFIG_PATH]

        config = cls()
        config.read_files(files)
        if cache_root is not None:
            config.cache_root = Path(cache_root)
        if expire_days is not None:
            config.expire_days = expire_days
        if auto_update is not None:
            config.auto_update = auto_update

        return config

    def read_files(self, files: list[str | PathLike]) -> None:
        parser = ConfigParser()
        parser.read([str(path) for path in files])
        self._update_from_configparser(parser)

    def _update_from_configparser(self, parser: ConfigParser) -> None:
        if parser.has_section("cache"):
            section = parser["cache"]
            if "root" in section:
                self.cache_root = Path(parser["cache"]["root"])
            if "expire_days" in section:
                self.expire_days = parser.getint("cache", "expire_days")
            if "auto_update" in section:
                self.auto_update = parser.getboolean("cache", "auto_update")
        if parser.has_section("ui"):
            section = parser["ui"]
            if "selector_command" in section:
                self.selector_command = section["selector_command"]



import argparse
from pathlib import Path

from minato.commands.subcommand import Subcommand
from minato.common import Selector
from minato.config import Config
from minato.minato import Minato


@Subcommand.register(
    name="update",
    description="update cached files",
)
class UpdateCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument(
            "query",
            nargs="*",
            default=[],
            help="query to filter cached files",
        )
        self.parser.add_argument(
            "--auto",
            action="store_true",
            help="check update and download files if updates are available",
        )
        self.parser.add_argument(
            "--force",
            action="store_true",
            help="update cached files without confirmation",
        )
        self.parser.add_argument(
            "--force-download",
            action="store_true",
            help="force to download files whether updates exist or not",
        )
        self.parser.add_argument(
            "--force-extract",
            action="store_true",
            help="force to extract archive files",
        )
        self.parser.add_argument(
            "--expired",
            action="store_true",
            help="update cached files that were expired",
        )
        self.parser.add_argument(
            "--failed",
            action="store_true",
            help="update cached files that were failed to download",
        )
        self.parser.add_argument(
            "--expire-days",
            type=int,
            default=None,
            help="specify expire days of the cache",
        )
        self.parser.add_argument(
            "--root",
            type=Path,
            default=None,
            help="specify the root directory path of cached data",
        )

    def run(self, args: argparse.Namespace) -> None:
        config = Config.load(
            cache_root=args.root,
        )
        minato = Minato(config)
        cache = minato.cache

        if args.query or args.expired or args.failed:
            cached_files = cache.filter(
                queries=args.query,
                expired=args.expired,
                failed=args.failed,
            )
        elif args.auto:
            cached_files = cache.all()
        else:
            candidates = {cached_file.url: cached_file for cached_file in cache.all()}
            selected_url = Selector(config.selector_command)(list(candidates.keys()))
            cached_files = [candidates[selected_url]] if selected_url in candidates else []

        cached_files = [
            x for x in cached_files if args.force_download or args.force_extract or minato.available_update(x.url)
        ]

        num_caches = len(cached_files)
        if num_caches == 0:
            print("No caches to update.")
            return

        print(f"{num_caches} files will be updated:")
        for cached_file in cached_files:
            print(f"  [{cached_file.uid[:8]}] {cached_file.url}")

        if not args.force:
            yes_or_no = input("Are you sure to update these caches? y/[n]:")
            if yes_or_no not in ("y", "Y"):
                print("canceled")
                return

        for cached_file in cached_files:
            minato.cached_path(
                cached_file.url,
                expire_days=args.expire_days,
                auto_update=args.auto,
                force_download=args.force_download,
                force_extract=args.force_extract,
            )

        print("Cache files were successfully updated.")



from __future__ import annotations

import os
import re
import sys
import time
from collections.abc import Sized
from typing import Any, Callable, Generic, Iterable, Iterator, TextIO, TypeVar, cast

T = TypeVar("T")
Self = TypeVar("Self", bound="Progress")

DISABLE_PROGRESSBAR = os.environ.get("MINATO_DISABLE_PROGRESSBAR", "0").lower() in ("1", "true")
ANSI_COLOR = re.compile(r"(\033|\x1b)\[[0-9;]*m")


def _dummy_iterator() -> Iterator[int]:
    iterations = 0
    while True:
        yield iterations
        iterations += 1


def _default_sizeof_formatter(size: int | float) -> str:
    if size % 1 < 1.0e-1:
        size = int(size)
    if isinstance(size, int):
        return str(size)
    return f"{size:.2f}"


def _truncate_text(text: str, width: int) -> str:
    if len(re.sub(ANSI_COLOR, "", text)) <= width:
        return text

    position = 0
    visible_length = 0
    truncated_text = ""
    for match in re.finditer(ANSI_COLOR, text):
        steps = min(match.start() - position, width - visible_length - 1)
        truncated_text += text[position : position + steps]
        truncated_text += text[match.start() : match.end()]
        visible_length += steps
        position = match.end()

    if visible_length < width:
        truncated_text += text[position : position + width - visible_length - 1]

    return truncated_text + "…"


class EMA:
    def __init__(
        self,
        alpha: float = 0.3,
    ) -> None:
        self._alpha = alpha
        self._value = 0.0

    def update(self, value: float) -> None:
        self._value = self._alpha * value + (1.0 - self._alpha) * self._value

    def get(self) -> float:
        return self._value

    def reset(self) -> None:
        self._value = 0.0


class Progress(Generic[T]):
    def __init__(
        self,
        total_or_iterable: int | Iterable[T] | None,
        desc: str | None = None,
        unit: str = "it",
        output: TextIO = sys.stderr,
        maxwidth: int | None = None,
        truncate: bool = True,
        partchars: str = " ▏▎▍▌▋▊▉█",
        framerate: float = 16.0,
        maxbarwidth: int | None = 40,
        sizeof_formatter: Callable[[int | float], str] = _default_sizeof_formatter,
        disable: bool = False,
    ) -> None:
        total_or_iterable = total_or_iterable or cast(Iterator[T], _dummy_iterator())
        self._iterable = (
            cast(Iterator[T], range(total_or_iterable)) if isinstance(total_or_iterable, int) else total_or_iterable
        )
        self._total = len(self._iterable) if isinstance(self._iterable, Sized) else None
        self._desc = desc
        self._unit = unit
        self._output = output
        self._maxwidth = maxwidth
        self._truncate = truncate
        self._partchars = partchars
        self._framerate = framerate
        self._maxbarwidth = maxbarwidth
        self._sizeof_formatter = sizeof_formatter
        self._disable = disable or DISABLE_PROGRESSBAR

        self._postfixes: dict[str, Any] = {}

        self._iterations = 0
        self._start_time = time.time()
        self._last_time = self._start_time
        self._last_time_rendered = self._start_time
        self._interval_ema = EMA()

    @staticmethod
    def _format_time(seconds: float) -> str:
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)
        if h > 0:
            return f"{int(h):d}:{int(m):02d}:{int(s):02d}"
        return f"{int(m):02d}:{int(s):02d}"

    def _get_maxwidth(self) -> int:
        try:
            terminal_width, _ = os.get_terminal_size()
        except OSError:
            terminal_width = 80
        if self._maxwidth:
            return min(terminal_width, self._maxwidth)
        return terminal_width

    def _get_bar(self, width: int, percentage: float) -> str:
        width = max(1, width)
        if self._maxbarwidth is not None:
            width = min(width, self._maxbarwidth)
        ratio = percentage / 100
        whole_width = int(ratio * width)
        part_width = int(len(self._partchars) * ((ratio * width) % 1))
        part_char = self._partchars[part_width]
        return f"{(self._partchars[-1] * whole_width + part_char)[:width]:{width}s}"

    def set_postfix(self, **postfixes: Any) -> None:
        self._postfixes = postfixes

    def show(self) -> None:
        if self._disable:
            return

        current_time = time.time()

        if 0 < self._iterations < (self._total or float("inf")):
            framerate = 1.0 / (current_time - self._last_time_rendered + 1.0e-13)
            if framerate > self._framerate:
                return

        template = ""
        contents: dict[str, Any] = {}

        elapsed_time = current_time - self._start_time
        interval_ema = self._interval_ema.get()
        average_iterations = 1.0 / interval_ema if interval_ema > 0.0 else 0.0

        contents["desc"] = self._desc
        contents["unit"] = self._unit
        contents["iterations"] = self._sizeof_formatter(self._iterations)
        contents["elapsed_time"] = self._format_time(elapsed_time)
        contents["average_iterations"] = self._sizeof_formatter(average_iterations)

        postfixes = [f"{key}={val}" for key, val in self._postfixes.items()]

        if self._desc:
            template = "{desc}: " + template
            contents["desc"] = self._desc

        if self._total is None:
            postfixes = [
                "{elapsed_time}",
                "{average_iterations}{unit}/s",
            ] + postfixes
            postfix_template = " ".join(postfixes)
            template = template + " {iterations}{unit} " + f"[{postfix_template}]"
        else:
            total_width = len(self._sizeof_formatter(self._total))
            percentage = 100 * self._iterations / self._total
            remaining_time = (self._total - self._iterations) * interval_ema

            postfixes = [
                "{elapsed_time}<{remaining_time}",
                "{average_iterations}{unit}/s",
            ] + postfixes
            postfix_template = " ".join(postfixes)

            template = (
                template + "{percentage:5.1f}% |{bar}| {iterations:>{total_width}}/{total} " + f"[{postfix_template}]"
            )

            contents["total_width"] = total_width
            contents["percentage"] = percentage
            contents["bar"] = ""
            contents["total"] = self._sizeof_formatter(self._total)
            contents["remaining_time"] = self._format_time(remaining_time)

            barwidth = max(1, self._get_maxwidth() - len(template.format(**contents)))
            contents["bar"] = self._get_bar(barwidth, percentage)

        line = template.format(**contents)
        if self._truncate:
            line = _truncate_text(line, self._get_maxwidth())

        self._output.write(f"\x1b[?25l\x1b[2K\r{line}")
        self._output.flush()

        self._last_time_rendered = current_time

    def update(self, iterations: int = 1) -> None:
        current_time = time.time()
        self._iterations += iterations
        self._last_time = current_time
        self._interval_ema.update((current_time - self._start_time) / self._iterations)
        self.show()

    def __iter__(self) -> Iterator[T]:
        self._iterations = 0
        self._start_time = time.time()

        with self:
            for item in self._iterable:
                yield item
                self.update()

    def __enter__(self: Self) -> Self:
        self._iterations = 0
        self._start_time = time.time()
        self.show()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.show()
        self._output.write("\x1b[?25h\n")
        self._output.flush()



import argparse
from pathlib import Path

from minato.commands.subcommand import Subcommand
from minato.common import Selector
from minato.config import Config
from minato.minato import Minato


@Subcommand.register(
    name="remove",
    description="remove cached files",
)
class RemoveCommand(Subcommand):
    def setup(self) -> None:
        self.parser.add_argument(
            "query",
            nargs="*",
            type=str,
            default=[],
            help="query to filter cached files",
        )
        self.parser.add_argument(
            "--force",
            action="store_true",
            help="remove files without confirmation",
        )
        self.parser.add_argument(
            "--expired",
            action="store_true",
            default=None,
            help="remove files that were expired",
        )
        self.parser.add_argument(
            "--failed",
            action="store_true",
            default=None,
            help="remove files that were failed to download",
        )
        self.parser.add_argument(
            "--root",
            type=Path,
            default=None,
            help="specify the root directory path of cached data",
        )

    def run(self, args: argparse.Namespace) -> None:
        config = Config.load(
            cache_root=args.root,
        )
        minato = Minato(config)
        cache = minato.cache

        if args.query or args.expired is not None or args.failed is not None:
            cached_files = cache.filter(
                queries=args.query,
                expired=args.expired,
                failed=args.failed,
            )
        else:
            candidtes = {cached_file.url: cached_file for cached_file in cache.all()}
            selected_url = Selector(config.selector_command)(list(candidtes.keys()))
            cached_files = [candidtes[selected_url]] if selected_url in candidtes else []

        num_caches = len(cached_files)

        if not cached_files:
            print("No files to delete")
            return

        print(f"{num_caches} files will be deleted:")
        for cached_file in cached_files:
            print(f"  [{cached_file.uid[:8]}] {cached_file.url}")

        if not args.force:
            yes_or_not = input("Delete these caches? y/[n]: ")
            if yes_or_not not in ("y", "Y"):
                print("canceled")
                return

        for cached_file in cached_files:
            with cache.lock(cached_file):
                cache.delete(cached_file)

        print("Cache files were successfully deleted.")



from __future__ import annotations

import dataclasses
import datetime
import hashlib
import json
import logging
import os
import uuid
from contextlib import contextmanager
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union

from minato.common import FileLock
from minato.exceptions import CacheAlreadyExists, CacheNotFoundError, ConfigurationError
from minato.util import remove_file_or_directory

logger = logging.getLogger(__name__)


class CacheStatus(Enum):
    PENDING = "PENDING"
    DOWNLOADING = "DOWNLOADING"
    EXTRACTING = "EXTRACTING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    DELETED = "DELETED"


@dataclasses.dataclass
class CachedFile:
    uid: str
    url: str
    local_path: Path
    created_at: datetime.datetime
    updated_at: datetime.datetime
    expire_days: int
    extraction_path: Optional[Path]
    status: CacheStatus
    version: Optional[str]
    auto_update: bool

    def __init__(
        self,
        uid: str,
        url: str,
        local_path: Union[str, PathLike],
        created_at: Union[str, datetime.datetime],
        updated_at: Union[str, datetime.datetime],
        expire_days: int = -1,
        extraction_path: Optional[Union[str, PathLike]] = None,
        status: Union[str, CacheStatus] = CacheStatus.PENDING,
        version: Optional[str] = None,
        auto_update: bool = True,
    ) -> None:
        if not isinstance(local_path, Path):
            local_path = Path(local_path)
        if isinstance(created_at, str):
            created_at = datetime.datetime.fromisoformat(created_at)
        if isinstance(updated_at, str):
            updated_at = datetime.datetime.fromisoformat(updated_at)
        if extraction_path is not None:
            if not isinstance(extraction_path, Path):
                extraction_path = Path(extraction_path)
            extraction_path = extraction_path.absolute()
        if isinstance(status, str):
            status = CacheStatus(status)

        self.uid = uid
        self.url = url
        self.local_path = local_path.absolute()
        self.created_at = created_at
        self.updated_at = updated_at
        self.expire_days = expire_days
        self.extraction_path = extraction_path
        self.status = status
        self.version = version
        self.auto_update = auto_update

    def to_dict(self) -> Dict[str, Any]:
        return {
            "uid": str(self.uid),
            "url": str(self.url),
            "local_path": str(self.local_path),
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
            "expire_days": self.expire_days,
            "extraction_path": str(self.extraction_path) if self.extraction_path else None,
            "status": self.status.value,
            "version": self.version,
            "auto_update": self.auto_update,
        }


class Cache:
    def __init__(
        self,
        root: Union[str, PathLike],
        default_expire_days: int = -1,
        default_auto_update: bool = True,
    ) -> None:
        root = Path(root)

        if not root.exists():
            os.makedirs(root, exist_ok=True)

        if not root.is_dir():
            raise ConfigurationError(f"Given cache_directory path is not a directory: {root}")

        self._root = root
        self._default_expire_days = default_expire_days
        self._default_auto_update = default_auto_update

    def __contains__(self, url: str) -> bool:
        try:
            self.by_url(url)
            return True
        except CacheNotFoundError:
            return False

    @contextmanager
    def lock(self, item: CachedFile) -> Iterator[None]:
        lock = FileLock(self.get_lockfile_path(item.uid))
        try:
            logger.debug("Trying to acquire file lock of %s.", item.url)
            lock.acquire()
            yield
        finally:
            lock.release()
            logger.debug("File lock of %s was released.", item.url)

    @staticmethod
    def _generate_uid(url: str) -> str:
        hashval = hashlib.md5(url.encode()).hexdigest()
        uuidval = uuid.uuid4().hex
        return f"{hashval}-{uuidval}"

    def get_metadata_path(self, uid: str) -> Path:
        return self._root / (uid + ".json")

    def get_lockfile_path(self, uid: str) -> Path:
        return self._root / (uid + ".lock")

    def load_cached_file(self, metadata_path: Path) -> CachedFile:
        if not metadata_path.exists():
            raise CacheNotFoundError(f"Cache not found: {metadata_path}")
        with metadata_path.open("r") as fp:
            params = json.load(fp)
        return CachedFile(**params)

    def new(self, url: str) -> CachedFile:
        uid = self._generate_uid(url)
        cached_file = CachedFile(
            uid=uid,
            url=url,
            local_path=self._root / uid,
            created_at=datetime.datetime.now(),
            updated_at=datetime.datetime.now(),
            expire_days=self._default_expire_days,
            auto_update=self._default_auto_update,
        )
        return cached_file

    def exists(self, item: CachedFile) -> bool:
        metadata_path = self.get_metadata_path(item.uid)
        return metadata_path.exists()

    def save(self, item: CachedFile) -> None:
        metadata_path = self.get_metadata_path(item.uid)
        with metadata_path.open("w") as fp:
            json.dump(item.to_dict(), fp)

    def add(self, item: CachedFile) -> CachedFile:
        if self.exists(item):
            raise CacheAlreadyExists(item.url)
        self.save(item)
        logger.debug("New cached file of %s was added.", item.url)
        return item

    def update(self, item: CachedFile) -> None:
        metadata_path = self.get_metadata_path(item.uid)
        if not metadata_path.exists():
            raise CacheNotFoundError(f"Cache not found with uid={item.uid}")
        item.updated_at = datetime.datetime.now()
        self.save(item)

    def by_uid(self, uid: str) -> CachedFile:
        metadata_path = self.get_metadata_path(uid)
        if not metadata_path.exists():
            raise CacheNotFoundError(f"Cache not found with uid={uid}")
        with metadata_path.open("r") as fp:
            params = json.load(fp)
        return CachedFile(**params)

    def by_url(self, url: str) -> CachedFile:
        logger.debug("Try to find cached file of %s", url)
        hashval = hashlib.md5(url.encode()).hexdigest()
        for metadata_path in self._root.glob(f"{hashval}-*.json"):
            cached_file = self.load_cached_file(metadata_path)
            if cached_file.url == url:
                logger.debug("Find cached file of %s: %s", url, cached_file.local_path)
                return cached_file

        logger.debug(
            "There is no cached files with hashval of %s, " "so try to find a corresponding file from all files in %s",
            url,
            self._root,
        )

        for cached_file in self.all():
            if cached_file.url == url:
                return cached_file
        raise CacheNotFoundError(f"Cache not found with url={url}")

    def delete(self, item: CachedFile) -> None:
        metadata_path = self.get_metadata_path(item.uid)
        lockfile_path = self.get_lockfile_path(item.uid)

        remove_file_or_directory(item.local_path)
        if item.extraction_path is not None:
            remove_file_or_directory(item.extraction_path)

        remove_file_or_directory(metadata_path)
        remove_file_or_directory(lockfile_path)
        logger.debug("A cached file of %s was successfully deleted.", item.url)

    def is_expired(self, item: CachedFile) -> bool:
        if item.expire_days < 0:
            return False
        now = datetime.datetime.now()
        delta = now - item.updated_at
        return delta.days >= item.expire_days

    def all(self) -> List[CachedFile]:
        cached_files: List[CachedFile] = []
        for metafile_path in self._root.glob("*.json"):
            with metafile_path.open("r") as fp:
                params = json.load(fp)
                cached_file = CachedFile(**params)
            cached_files.append(cached_file)
        cached_files = sorted(cached_files, key=lambda x: x.created_at)
        return cached_files

    def filter(
        self,
        queries: List[str],
        expired: Optional[bool] = None,
        failed: Optional[bool] = None,
        completed: Optional[bool] = None,
    ) -> List[CachedFile]:
        cached_files = self.all()
        for query in queries:
            cached_files = [x for x in cached_files if query in x.url or x.uid.startswith(query)]
        if expired is not None:
            cached_files = [x for x in cached_files if self.is_expired(x) == expired]
        if failed is not None:
            cached_files = [x for x in cached_files if (x.status == CacheStatus.FAILED) == failed]
        if completed is not None:
            cached_files = [x for x in cached_files if (x.status == CacheStatus.COMPLETED) == completed]

        unique_caches = {x.uid: x for x in cached_files}
        cached_files = sorted(unique_caches.values(), key=lambda x: x.created_at)
        return cached_files



from __future__ import annotations

import os
import re
import sys
import unicodedata
from typing import Any, ClassVar, TextIO

CSI = "\x1b"
REGEX_ANSI_CSI = re.compile(rf"{CSI}\[[0-9]+[a-zA-Z]")


class Table:
    MIN_COLUMN_WIDTH: ClassVar = 2

    def __init__(self, columns: list[str], shrink: bool = True) -> None:
        self._columns = columns
        self._items: list[dict[str, Any]] = []
        self._shrink = shrink

    def __getitem__(self, columns: list[str]) -> "Table":
        table = Table(columns=columns)
        for item in self._items:
            table.add(item)
        return table

    @staticmethod
    def _get_column_value_str(value: Any) -> str:
        return str(value)

    @staticmethod
    def _remove_ansi_code(value: str) -> str:
        value = re.sub(REGEX_ANSI_CSI, "", value)
        return value

    @staticmethod
    def _is_fullwidth(character: str) -> int:
        return unicodedata.east_asian_width(character) in "AFW"

    def _get_str_width(self, value: str) -> int:
        value = self._remove_ansi_code(value)
        return sum(2 if self._is_fullwidth(c) else 1 for c in value)

    def _get_padded_column_value(self, value: str, width: int) -> str:
        width = max(width, Table.MIN_COLUMN_WIDTH)
        value_width = self._get_str_width(value)
        if value_width > width:
            current_value = ""
            current_width = 0
            current_index = 0
            while True:
                character = value[current_index]
                character_width = self._get_str_width(character)
                if current_width + character_width < width:
                    current_value += character
                    current_width += character_width
                    current_index += 1
                else:
                    break
            value = current_value + "\u2026"
            value_width = current_width + 1
        return value + " " * max(0, width - value_width)

    def _get_column_widths(self) -> dict[str, int]:
        column_widths: dict[str, int] = {}
        for col in self.columns:
            column_values = [x[col] for x in self._items]
            column_value_strings = [self._get_column_value_str(x) for x in column_values]

            column_width = max(self._get_str_width(x) for x in column_value_strings + [col])

            column_widths[col] = column_width

        if not self._shrink:
            return column_widths

        num_columns = len(self.columns)
        terminal_width, _ = os.get_terminal_size()

        total_column_width = num_columns + sum(column_widths.values()) - 1
        previous_width = -1
        while total_column_width > terminal_width and previous_width != total_column_width:
            width, column = max((width, column) for column, width in column_widths.items())
            column_widths[column] = max(width - 1, Table.MIN_COLUMN_WIDTH)
            previous_width = total_column_width
            total_column_width = num_columns + sum(column_widths.values()) - 1
        return column_widths

    @property
    def columns(self) -> list[str]:
        return self._columns

    def add(self, item: dict[str, Any]) -> None:
        self._items.append({col: item[col] for col in self.columns})

    def sort(self, column: str, desc: bool = False) -> None:
        def _key(item: dict[str, Any]) -> Any:
            return item[column]

        self._items = sorted(self._items, key=_key, reverse=desc)

    def filter(self, query: str | dict[str, str]) -> "Table":
        if isinstance(query, str):
            query = {col: query for col in self.columns}
        table = Table(columns=self.columns)
        for item in self._items:
            for c, q in query.items():
                if q in item[c]:
                    table.add(item)
                    break
        return table

    def show(self, output: TextIO | None = None) -> None:
        if output is None:
            output = sys.stdout

        column_widths = self._get_column_widths()

        output.write(" ".join(self._get_padded_column_value(col, column_widths[col]) for col in self.columns) + "\n")
        output.write(" ".join("=" * column_widths[col] for col in self.columns) + "\n")
        for item in self._items:
            output.write(
                " ".join(
                    self._get_padded_column_value(self._get_column_value_str(item[col]), column_widths[col])
                    for col in self.columns
                )
                + "\n"
            )
